class APIKeyRequiredError(BaseException):
    pass


class ClientStringUnacceptableError(BaseException):
    pass


class ClientListUnacceptableError(BaseException):
    pass


class ConfItemNotFound(BaseException):
    pass


class AllClientsFailedError(BaseException):
    pass


class InvalidKeyTypeError(BaseException):
    pass


# Builtin imports
import logging
import os
import json

# Internal imports
from smartsearch import exceptions, clients

# External imports
import nltk

_client_classes = {
    "wolfram": clients.Wolfram,
    "google": clients.Google
}

log = logging.getLogger()

# The required nltk corpora, from newspaper
REQUIRED_CORPORA = [
    'brown',  # Required for FastNPExtractor
    'punkt',  # Required for WordTokenizer
    'maxent_treebank_pos_tagger',  # Required for NLTKTagger
    'movie_reviews',  # Required for NaiveBayesAnalyzer
    'wordnet',  # Required for lemmatization and Wordnet
    'stopwords'
]

# Generate a list of all the modules that still need to be downloaded
not_downloaded = []

# The downloaded nltk modules
found = []
nltk_dirs = [
    "corpora",
    "tokenizers",
    "chunkers",
    "taggers"
]
for d in nltk_dirs:
    try:
        found += os.listdir(nltk.data.find(d))
    except LookupError:
        pass

for corpus in REQUIRED_CORPORA:
    corpus_name = "{}.zip".format(corpus)
    if corpus_name not in found:
        not_downloaded.append(corpus)


def nltk_download():
    """
    Download missing nltk modules required for the nlp functions in newspaper
    """
    log.info("{0} nltk modules required by the newspaper module have not been downloaded, downloading "
             "now.".format(len(not_downloaded)))
    for mod in not_downloaded:
        log.info("Downloading {}".format(mod))
        nltk.download(mod)


class Searcher:
    clients = {}
    keys = {}
    query_order = [
        "wolfram",
        "google"
    ]

    def __init__(self, keys=None, conf_file=None):
        """
        Instantiate a searcher object

        :param keys: A dictionary of keys and usage information for search clients
        :param conf_file: A configuration file to optionally load keys from
        """
        # Download the nltk modules required by newspaper that haven't been downloaded yet
        if not_downloaded:
            nltk_download()
        if conf_file:
            if os.path.isfile(conf_file):
                j = json.load(open(conf_file))
                self.keys = j
        else:
            self.keys = keys
        # Validate the API keys
        if keys:
            for key, key_data in self.keys.items():
                if key in _client_classes.keys():
                    if type(key_data) != str:
                        raise exceptions.InvalidKeyTypeError("Keys must be of type str")
        else:
            log.warning("No keys found")
        for client_name, client_baseclass in _client_classes.items():
            if client_baseclass.key_required:
                # If the client name is included in the loaded keys, instantiate the client with the key data.
                # If it's not included, warn in the logs that the client will be skipped
                if self.keys and client_name in self.keys.keys():
                    # Instantiate the client with the key information
                    instantiated_client = client_baseclass(self.keys[client_name])
                    self.clients.update({client_name: instantiated_client})
                    log.debug("Loaded client {0} with provided key information.".format(client_name))
                else:
                    log.warning("Skipping client {0} since it requires an api key that was not passed.".format(client_name))
            else:
                instantiated_client = client_baseclass()
                self.clients.update({client_name: instantiated_client})
                log.debug("Loaded client {0}.".format(client_name))

    @property
    def usages(self):
        usage_data = {}
        for client, client_class in self.clients.items():
            if client_class.key_required:
                usage_data.update(client.usages)
        return usage_data

    def query(self, query_text):
        for client in self.query_order:
            if client in self.clients.keys():
                resp = self.clients[client].query(query_text)
                if resp:
                    return resp
        # Raise an error if none of the clients found a result or completed without error
        raise exceptions.AllClientsFailedError("None of the enabled clients {0} could find an answer to the query "
                                               "{1}".format(", ".join(self.clients.keys()), query_text))

    def save(self, filename="search.conf"):
        """
        Save the current configuration to a file

        :param filename: The file to save it to. If none is specified, saved as "search.conf"
        """
        if self.keys:
            log.info("Caching search settings in file {}".format(filename))
            with open(filename, "w") as f:
                json_data = json.dumps(self.keys)
                f.write(json_data)
        else:
            log.warning("No configuration to cache")




from setuptools import setup
from pip.req import parse_requirements
import uuid

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements("requirements.txt", session=uuid.uuid1())

# reqs is a list of requirement
# e.g. ['django==1.5.1', 'mezzanine==1.4.6']
reqs = [str(ir.req) for ir in install_reqs]

setup(name='smartsearch',
      version='1.0.5',
      description='A module that searches various internet sources to find direct answers to queries.',
      url='https://github.com/ironman5366/smartsearch',
      author='Will Beddow',
      author_email='will@willbeddow.com',
      license='MIT',
      packages=['smartsearch'],
      zip_safe=False, install_requires=reqs)




# Builtin imports
import logging

#  Internal imports
from smartsearch import exceptions

# External imports
import wolframalpha
import google
import wikipedia
from newspaper import Article
import requests
from bs4 import BeautifulSoup


log = logging.getLogger()


class Client:
    """
    Basic query client
    """
    client_name = None
    key_required = False

    def __init__(self, key=None):
        """
        Instantiate a client

        :param key: The API key for the client
        """
        if self.key_required:
            if not key:
                raise exceptions.APIKeyRequiredError("Client {} must have an API key to run queries".format(self.client_name))
        self._key = key
        self._build_instance()

    def _build_instance(self):
        """
        Build the instance of the query
        """
        pass

    def query(self, query_text):
        """
        Send a query to the client

        :param query_text: The text of the query
        :return response: The response to the query
        """
        raise NotImplementedError("Client class should nto be called directly")


class Wolfram(Client):
    _client_instance = None
    key_required = True

    def _build_instance(self):
        """
        Instantiate wolframalpha.Client
        """
        self._client_instance = wolframalpha.Client(self._key)

    def query(self, query_text):
        """
        Use the instantiated client to query
        """
        res = self._client_instance.query(query_text)
        try:
            next_result = res.results.__next__().text
            if next_result:
                return next_result
            else:
                return False
        except (StopIteration, AttributeError) as e:
            return False


class Google(Client):
    _key_required = False

    def query(self, query_text):
        # Query google
        search_object = google.search(query_text)
        # Determine if a wikipedia url is in the first 5 searches
        urls = []
        for i in range(0, 4):
            url = search_object.__next__()
            urls.append(url)
            # If it is, query with wikipedia
            if "wikipedia.org/wiki" in url:
                # Until I can find a better way to do this, I'm grabbing the title from the url, and querying that way
                article_title = url.split("wikipedia.org/wiki/")[1].replace("_", " ")
                log.debug("Found wikipedia article {0}".format(article_title))
                response = wikipedia.summary(article_title, sentences=2) + " ({0})".format(url)
                return response
        # If there were no wikipedia pages
        first_url = urls[0]
        try:
            article = Article(first_url)
            article.download()
            article.parse()
            article.nlp()
            article_summary = article.summary
            article_title = article.title
            return "{0}\n{1} - ({2})".format(
                article_summary, article_title, first_url
            )

        except Exception as article_exception:
            try:
                log.debug("Got error {0} while using newspaper, switching to bs4".format(
                    article_exception.args
                ))
                html = requests.get(first_url).text
                # Parse the html using bs4
                soup = BeautifulSoup(html, "html.parser")
                [s.extract() for s in soup(['style', 'script', '[document]', 'head', 'title'])]
                text = soup.getText()
                # break into lines and remove leading and trailing space on each
                lines = (line.strip() for line in text.splitlines())
                # break multi-headlines into a line each
                chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
                # drop blank lines
                soup_text = '\n'.join(chunk for chunk in chunks if " " in chunk)
                response = format(soup_text) + " ({0})".format(first_url)
                return response
            except Exception as search_exception:
                log.error("Error {0} occurred while searching query {1}".format(
                   search_exception.args, query_text
                ))
                return False
